//! Rust Challenge

use std::future::Future;
use std::pin::Pin;

use tokio::io::AsyncWriteExt;
use tokio::sync::{mpsc, watch};
use tokio::time::{Duration, Interval};

/// Just an easier to read type.
type IoResult<T> = Result<T, std::io::Error>;

/// The commands available to send to the Satellite.
#[derive(Clone, Copy)]
pub enum Command
{
   /// Fire the engine in T minus x seconds!
   Propulsion
   {
      /// The amount of seconds to wait before we fire the engines.
      delay: Duration
   },

   /// Cancel any waiting commands.
   Cancel
}

/// The state of the program.
#[derive(Clone, Copy)]
pub enum ProgramState
{
   /// The program is currently running.
   Running,

   /// The program has switched to shutting down.
   Terminated
}

/// Reads input from the user until the termination signal is
/// received. It will send command messages every time it reads valid input.
fn read_user_input(command_sender: mpsc::Sender<Command>,
                   term_sender: watch::Sender<ProgramState>)
                   -> IoResult<()>
{
   let mut running: bool = true;
   let mut buffer = String::new();

   // Loop and read the user input.
   while running
   {
      // Clearing the last read line and read a new one.
      buffer.clear();
      std::io::stdin().read_line(&mut buffer)?;

      let input = buffer.trim();

      match input
      {
         // A termination method wasn't specified in the document, so
         // this will terminate when either stop, exit, or quit is entered.
         "exit" | "stop" | "quit" =>
         {
            println!("Exiting program.");
            running = false;
            term_sender.send_replace(ProgramState::Terminated);
         }

         _ =>
         {
            // Check to see if we were given a number.
            match input.parse::<i32>()
            {
               Ok(seconds) =>
               {
                  // Here we are just handling the different possible numbers
                  // passed in. -1 cancels commands, less than -1 is ignored
                  // with a message, and greated than -1 is turned into a
                  // propulsion command.
                  if seconds == -1
                  {
                     match command_sender.blocking_send(Command::Cancel)
                     {
                        Ok(_) =>
                        {}
                        Err(e) => eprintln!("Failed to send command: {}", e)
                     }
                  }
                  else if seconds < -1
                  {
                     println!("All propulsion delay times are given in \
                               seconds within the range of [{}, {}]",
                              0,
                              i32::MAX);
                     println!("A value of -1 will cancel any current \
                               propulsion commands.");
                  }
                  else
                  {
                     // Time to create the Propulsion command.
                     let delay_secs: u64 = seconds as u64;
                     let delay_duration: Duration =
                        Duration::from_secs(delay_secs);
                     match command_sender.try_send(Command::Propulsion {
                        delay: delay_duration,
                     }) {
                        Ok(_) => {}
                        Err(e) => eprintln!("Failed to send command: {}", e),
                     }
                  }
               }

               Err(e) =>
               {
                  println!("Unable to parse the given input into seconds.");
                  println!("{:?}\n", e);
                  println!("Please specify only the desired whole seconds \
                            until the time to fire the engines.");
               }
            }
         }
      }
   }

   Ok(())
}

/// Determines if there is an interval that needs to be run or if
/// there is nothing to really do.
///
/// This was the trickiest part due to the syntax.
fn maybe_tick<'a>(interval: Option<&'a mut Interval>)
                  -> Pin<Box<dyn Future<Output = ()> + Send + 'a>>
{
   match interval
   {
      Some(interval) =>
      {
         Box::pin(async move {
            interval.tick().await;
            ()
         })
      }

      None => Box::pin(std::future::pending())
   }
}

/// The command task.
///
/// Processes the commands sent and fire the engine if a given
/// propulsion command completes.
async fn process_commands(mut command_receiver: mpsc::Receiver<Command>,
                          mut term_receiver: watch::Receiver<ProgramState>)
                          -> IoResult<()>
{
   let mut stdout = tokio::io::stdout();
   let mut propulsion_interval: Option<Interval> = None;
   let mut running: bool = true;

   while running
   {
      tokio::select! {
          // Process commands.
          Some(command) = command_receiver.recv() =>
          {
             match command
             {
                Command::Cancel =>
                {
                   // Dump any interval that is running.
                   propulsion_interval = None;
                }

                Command::Propulsion { delay } =>
                {
                   // Overwrite any interval with this great new interval.
                   propulsion_interval = Some(tokio::time::interval(delay));

                   // Skip the first immediate tick.
                   if let Some(interval) = propulsion_interval.as_mut()
                   {
                      interval.tick().await;
                   }
                }
             }
          }

          // Run the interval timer if there is one.
          _ = maybe_tick(propulsion_interval.as_mut()) =>
          {
            stdout.write_all(b"firing now!\n").await?;
            stdout.flush().await?;

            // Clear the command out since it is done.
            propulsion_interval = None;
          }

          // Listen for termination.
          _ = term_receiver.changed() =>
          {
              running = match *term_receiver.borrow_and_update()
              {
                  ProgramState::Running => { true }
                  ProgramState::Terminated => { false }
              };
          }
      }
   }

   Ok(())
}

/// Program entry point.
#[tokio::main]
async fn main() -> IoResult<()>
{
   // The channel that will be used to send satellite commands between tasks.
   let (command_sender, command_receiver) = mpsc::channel::<Command>(100);

   // The channel that will be used to signal termination of the program.
   //
   // This could be done as a satellite command since the program is driven by
   // user interaction and it all happens from the input_task, but this is a
   // program signal so I prefer to keep it seperate from the command messages.
   // It also allows for moving termination control to another task, say from
   // Ctrl+C, if the program was changed to use TCP or something else.
   let (term_sender, term_receiver) = watch::channel(ProgramState::Running);

   // Spawn a task to handle reading input from the user. We use a thread here
   // because tokio recommends it for the blocking read calls for interactive
   // user input.
   let input_task = tokio::task::spawn_blocking(move || {
      read_user_input(command_sender, term_sender)
   });

   // Spawn a task to handle simulating the satellite.
   let command_task = tokio::spawn(async move {
      process_commands(command_receiver, term_receiver).await
   });

   // Run both tasks at the same time.
   let (command_result, input_result) = tokio::join!(command_task, input_task);

   // Handle errors from the tasks.
   match command_result
   {
      Ok(Ok(())) =>
      {}

      Ok(Err(e)) =>
      {
         eprintln!("Command task failed: {}", e);
         return Err(e);
      }

      Err(e) =>
      {
         eprintln!("Command task panicked or was aborted: {}", e);
         return Err(std::io::Error::new(std::io::ErrorKind::Other, e));
      }
   }

   match input_result
   {
      Ok(Ok(())) =>
      {}

      Ok(Err(e)) =>
      {
         eprintln!("Input task failed: {}", e);
         return Err(e);
      }

      Err(e) =>
      {
         eprintln!("Input task panicked or was aborted: {}", e);
         return Err(std::io::Error::new(std::io::ErrorKind::Other, e));
      }
   }

   println!("Shutdown complete.");
   Ok(())
}
